/*******************************************************************************
 Program name:     lm_bayes
                   (Version 2.6: MORGAN compatible)
                   ----------------------------
  Author:          Andrew W. George

  Date:            5th February, 2002

  Keywords:        Pseudo-Bayesian inference, lod scores, LM-samplers,
                   meiosis indicators, sequential imputation, 
                   Metropolis-Hastings restarts, 
                   Rao-Blackwellization estimators.

  Purpose:         To calculate multipoint lod scores from data observed on 
                   general pedigrees using a MCMC method which is less 
                   susceptible to mixing problems.

  Methods
  Utilized:        LM samplers, M-H and Gibbs samplers, sequential imputation 
                   for generating starting marker configurations,  
                   Metropolis-Hastings restarts via sequential imputation, 
                   Rao-Blackwellization estimator of the likelihood ratio.

  Potential
  Improvements:    More realistic penetrance function, improved sequential 
                   imputation, handling unknown genetic models

  Acknowledgement: I am indebted to the many people who under Elizabeth 
                   Thompson's mentorship, have produced a suite of library 
                   modules known as MORGAN. I have reaped the benefits of many 
                   years of development and experimentation. I would also like
                   to thank Myrna Jewett who has patiently answered my many 
                   coding questions.

  Comments:        ability to handle binary trait with liability classes 
                                                                    AWG 08/23/02

*******************************************************************************/

/**
 * @addtogroup lm_bayes
 * @{
 */

#define PNAM        "lm_bayes"
#define LIABILITY   0                   /* reset to nonzero value if there are
                                           liability classes                  */
/* MAKE SURE THE CORRECT penetrance function is set by allocate_loci()        */
/* set trait data phenotypic in parameter file -- e.g. bayes.par in Gold2     */

/* general included files                                                     */
#include "autozyg.h"
#include "lm_bayes.h"

/* output file handling include files                                         */
#include "./NewRtnes/write_rb.c"
#include "./NewRtnes/set_previous_realiz.c"
#include "./NewRtnes/calc_raoI.c"

/* M-H proposals                                                              */
#include "./NewRtnes/update_unlnkd_2_lnkd.c"
#include "./NewRtnes/update_loci.c"
#include "./NewRtnes/update_position.c"

/* MH restart                                                                 */
#include "./NewRtnes/mh_restart.c"

/* Pseudo-prior calculation                                                   */
#include "./NewRtnes/norm_pseudo_prior.c"

/* free memory                                                                */
#include "./NewRtnes/free_local.c"

int   main (int argc, char *argv[])
{
   /* local variable declaration                                              */
   int   i, ii, k, c;                   /* indexes                            */
   int   cnt;                           /* counter                            */
   int   err;                           /* error variable for seg_pen2()      */
   int   ngenes = 0;                    /* number of genes in pedigree,
                                           initialize to 0                    */
   int   sx;                            /* sex, female or male                */
   int   nfou;                          /* number of founders                 */
   int   nmrkloc;                       /* number of marker loci              */
   int   nsw;                           /* total number of swiches in 
                                           pedigree file <= number of meioses */
   int   iteration;                     /* index iteration number of
                                           MCMC sampler                       */
   int  *nns;                           /* number of switches/component
                                           Note - array index begins at 1     */
   int   rao_freq = 0;                  /* count of iterations
                                           for the rao-Bl.. estimator         */
   int **dodo = NULL;
   int ***y;                            /* trait and marker phenotypes
                                           y[mat, pat][0,.., L][0,.., NObs]   */
   double *pseudo_prior_calc;           /* used in calculating appropriate
                                           pseudo_priors                      */
   double *rao_realiz;                  /* Rao-Blackwellize estimate over 
                                           all test positions for a specific
                                           realization                        */
   double **prob;                       /* required by the M sampler
                                           [num compnts][num loci]            */
   double **quant = NULL;               /* (in preparation) Will be eventually
                                           used with a quantitative trait     */
   double p_linked;                     /* Prob of selecting the proposal
                                           which moves the trait locus along
                                           the linkage group as opposed to
                                           moving the trait locus to
                                           an unlinked position               */
   FILE *pedF;                          /* file pntr to input pedigree file   */
   FILE *oSeedF;                        /* file pntr to seed file             */
   FILE *oRBF;                          /* file pntr to output file of
                                           rb estimates                       */
   FILE *fptr_pseudo_input = NULL;      /* file pntr to input  file for
                                           the pseudo-priors                  */

   /*  +-----------------------+  */
   /*  |   Begin program body  |  */
   /*  +=======================+  */

   atexit (print_tail);
   print_head (*argv);

   /* +-------------------------------------+                                 */
   /* |   PROCESSING OF THE INPUT           |                                 */
   /* |   PARAMETER STATEMENTS VIA          |                                 */
   /* |   THE PROC PROCEDURES               |                                 */
   /* +=====================================+                                 */

   if (!proc_pars (argc, argv, PNAM))   /* process input parameter      */
      exit (EXIT_FAILURE);              /* statements                   */

   if (!proc_auto ())                   /* check statements for         */
      exit (EXIT_FAILURE);              /* completeness and consistency */

   if (LIABILITY)
   {
      printf (" WARNING! Liability classes have been requested!! \n");
      printf (" Have you specified set trait data pheno? \n");
      printf (" Have you also changed the penetrance routine?\n\n");

      /* the expected order in the pedigree record is liability class,
         immediately followed by the trait phenotype.
         these are to be stored in the reverse order in data[][]
       */
      for (i = 1; i < (SexIn != N ? 0 : 1) + NInts; ++i)
         if (SelInt[i])
         {
            SelInt[i + 1] = SelInt[i];
            ++SelInt[i];
            ++NColData;
            ++i;
         }
   }
   if (!proc_seeds ())                  /* set up seed                  */
      exit (EXIT_FAILURE);              /* parameters                   */

   if (!proc_fils ())                   /* set up pedigree file         */
      exit (EXIT_FAILURE);              /* parameters                   */

   if (!chk_pedfil ())                  /* checks pedigree file for     */
      give_up (ERR_qit);                /* errors.                      */

   if (!mak_opedfil ())                 /* writes new pedigree file     */
      give_up (ERR_qit);                /* if changes                   */

   if (!(pedF = open_ped ()))           /*opens pedigree file           */
      give_up (ERR_qit);

   nmrkloc = NLoci - 1;
   p_linked = .5;

   /*  +-------------------------------------+  */
   /*  |   ALLOCATION AND INITIALIZATION     +  */
   /*  +=====================================+  */

   alloc_ped ();
   input_ped (pedF, 0, dodo, quant);    /* input pedigree and trait data */

   nfou = count_founders ();

   MaxNsw = 2 * (pedsize - nfou);
   nns = malloc ((n_compnts + 1) * sizeof (int));
   alloc_chk (WHERE, nns);

   y = alloc_y (nmrkloc, pedsize);

   /* set marker data for observed individuals                             */
   set_markdat (y, nmrkloc, n_compnts);

   /* allocate substructures associated with the global variable
    * previous_realiz */
   alloc_previous_realiz (pedsize, nmrkloc);

   /* qdex and trtphen are in preparation  for quantitative traits            */
   if (TraitPheno)
   {
      qdex = calloc ((pedsize + 1), sizeof (int));
      alloc_chk (WHERE, qdex);
      trtphen = calloc ((pedsize + 1), sizeof (int));
      alloc_chk (WHERE, qdex);
   }

   /* set and print trait data                                                */
   set_traitdat (y, qdex, trtphen, pedsize, n_compnts);
   prnt_traitdat (y, pedsize, n_compnts);

   /* set and index switches                                                  */
   alloc_xxgg (nmrkloc, pedsize);

   /* initialize seeds for rans() */
   setSeeds (SAMPL);

   nsw = list_switches (nmrkloc, nns);
   printf (" Number of meiosis switches to sample is %d\n", nsw);
   fflush (stdout);

   /* allocate pseudo-prior array                                             */
   if (!InPseudo)
   {
      PseudoPrior = calloc (NTPos1, sizeof (double));  /* array begins at 0   */
      alloc_chk (WHERE, PseudoPrior);   /* - unlinked as well                 */

      /* Set pseudo-prior to unnormalized uniform */
      for (ii = 0; ii < NTPos1; ii++)
         PseudoPrior[ii] = 1.0;

      /* allocate pseudo-prior-calc array                                     */
      pseudo_prior_calc = calloc (NTPos1, sizeof (double));  /* array begins  */
      alloc_chk (WHERE, pseudo_prior_calc);  /* at 0, unlinked as well        */
   }

   /* allocate rao_realiz array                                               */
   rao_realiz = calloc (NTPos1, sizeof (double));  /* array begins at 0,
                                                      unlinked as well        */
   alloc_chk (WHERE, rao_realiz);

   /* allocate rao blackwellization estimator array                           */
   raoI_est = calloc (NTPos1, sizeof (double));  /* array begins at 0 */

   /* allocate probability distributions for  markers                         */
   mrk_distr = malloc ((nmrkloc + 1) * sizeof (double));  /* arry begins at 1 */
   alloc_chk (WHERE, mrk_distr);

   /* allocate array for frequency MCMC sampler visits each test position     */
   tposit_freq = calloc (NTPos1, sizeof (int));  /* array begins at 0  */
   alloc_chk (WHERE, tposit_freq);

   /* allocate arrays for M sampler */
   alloc_msmpl ();

   /*   +-----------------------------------------------+  */
   /*   |   ASSIGNING STARTING VALUES TO PARAMETERS     |  */
   /*   +===============================================+  */

   /*   Idea:  Using sequential imputation,                                   *
    *          generate S_j^* and associated weights w_j                      *
    *          (i.e. just single locus peeling).                              *
    *          Then S^* and w^* are easily obtained.                          *
    *          Choose an S^* as a starting value                              *
    *          with prob \frac{w^*}{W} where W = sum w^*                      */

   alloc_chromo (nmrkloc, 0);

   for (sx = FEM; sx <= MAL; sx++)
      set_chromosomes (nmrkloc, Marek[sx], 0, &IntVal[0], &RecVal[sx][0], sx);

   /* Allocate locus structures    -moved to main program, EAT June 2003 */
   /* if TraitPheno either liability classes or incomplete penetrance    */
   /*  will be called :    EAT  12/2003                                  */

   if (LIABILITY)
      allocate_loci (nmrkloc + 1, NAll, Qq, pedF, 3);
   else
      allocate_loci (nmrkloc + 1, NAll, Qq, pedF, 1);

   /* Use locus-by-locus sampling or sequential imputation for setup and for
      initial xx0, xx1 (maternal and paternal segregation indicators) */

   if (SmplSetup == LSMPL)              /* use locus-by-locus sampling */
      set_lsmpl (0, nmrkloc, y, TraitPheno);
   else                                 /* use sequential imputation */
      seq_imp (NSeqImp, pedsize, nmrkloc + 1, y, (double *) 0, 1.0);

   /* Note - unlinked is interval 0, test position 0 */
   set_previous_realiz (0, 0, 0.5, 0.5, 0, xx0, xx1, nmrkloc, pedsize);

   for (i = k = 0; i <= nmrkloc; i++)
   {
      ngenes = set_genes (i);

      /* Allocate memory for gen_pen() */

      if (loci[i].locus_type == TRAIT_LOCUS)
         alloc_gen_pen (ngenes, i);
      else if (!k++)
         alloc_seg_pen (ngenes, nmrkloc + 1, y, &prob);
   }
   for (i = 0; i <= nmrkloc; i++)
      for (c = 0; c < n_compnts; c++)
         prob[c][i] = seg_pen2 (c, y, Qq[i], i, ngenes, &err);

   prnt_mcmcstrt ();

   /*   +---------------------------------------+           */
   /*   |  BURN-IN                              |           */
   /*   +=======================================+           */

   /* PERFORM NBrnIt iterations (i.e. burn-in iterations) to allow the      */
   /*   S's to mix correctly. The trait locus remains unlinked.             */

   printf (" Beginning burn-in iterations \n");
   cnt = 1;                             /* Initialize counter */
   for (iteration = 1; iteration <= NBrnIt; iteration++)
   {
      if (!(iteration % ChkIt))
      {
         printf (" Starting burn-in iteration %d\n", iteration);
         fflush (stdout);
      }
      update_loci (LSamplProb, nmrkloc, pedsize, y, nsw, prob, ngenes);
   }
   printf ("\n");

   /*   +---------------------------------------+       */
   /*   |  PSUEDO-PRIOR ASSIGNMENT              |       */
   /*   +=======================================+       */

/* To allow the sampler to explore regions with low posterior probability   */
/* a pseudo-prior (according to E(log P(Y_T | S_M))^-1) is also created     */

   /* If the user has not specified pseudo-priors, calculate them */
   /* by running the MCMC method with a uniform pseudo-prior      */
   if (!InPseudo)
   {
      printf (" Beginning pseudo-prior calculations ........ \n");

      /* Initialize marker prob distribution to uniform. Change if
         one marker locus is to be favoured */
      for (ii = 1; ii <= nmrkloc; ii++)
         mrk_distr[ii] = 1.0 / nmrkloc;

      rao_freq = 0;

      for (iteration = 1; iteration <= NPreIt; iteration++)
      {
         if (!(iteration % ChkIt))
         {
            printf (" Starting preliminary iteration %d\n", iteration);
            fflush (stdout);
         }

         update_loci (LSamplProb, nmrkloc, pedsize, y, nsw, prob, ngenes);

         if (previous_realiz.intvl == 0 && previous_realiz.rec_lft[FEM] == 0.5)
            /* Trait locus is presently unlinked */
            update_unlnkd_2_lnkd (nmrkloc, pedsize, y, prob, ngenes, n_compnts,
                                  p_linked);

         /* Trait locus is presenlty linked */
         else                           /* Trait locus is presenlty linked */
            update_position (p_linked, WindWdth, nmrkloc, pedsize, y, prob,
                             ngenes, n_compnts);

         if (!(iteration % ScorIt))     /* Calculate Rao-Bl estimator */
         {
            rao_freq++;
            calc_raoI (nmrkloc, rao_freq, rao_realiz);

            /* Using rao blackwellize estimator to calc pseudo-priors         */
            for (ii = 0; ii < NTPos1; ii++)  /* incremental mean. Saves us from
                                                having to store all the lange
                                                estimates                      */

               /* -1.0 appears below because we are taking
                  inverse log P(Y_T | S_M)  */
               pseudo_prior_calc[ii] = (rao_freq - 1) * pseudo_prior_calc[ii]
                  / rao_freq + (rao_realiz[ii]) / rao_freq;
         }

         if (SeqImpIt && !(iteration % SeqImpIt))
            mh_restart (nmrkloc, pedsize, prob, y, n_compnts, ngenes);

      }                                 /* END LOOP */
      printf ("\n");

      normalize_pseudo_prior (pseudo_prior_calc);  /*  normalize              */

      /* Reset pseudo priors from uniform to the values based upon the
         rao-blackwellization estimates */
      for (ii = 0; ii < NTPos1; ii++)
         PseudoPrior[ii] = pseudo_prior_calc[ii];
   }

   /* open output R-B estimates file, if requested */

   if (OutRB)
      if (!(oRBF = open_fil (ORBEST)))
         give_up (ERR_qit);

   /*  +------------------------+                                             */
   /*  |  BEGIN MCMC SAMPLER    |                                             */
   /*  +========================+                                             */

   printf (" Beginning main MCMC sampler \n");

#if 0                                   /* (pseudo-priors are printed later */
   printf (" \n\n ");
   printf ("  PSEUDO-PRIORS\n\n");
   for (ii = 0; ii < NTPos1; ii++)
      printf ("pseudo_prb[%2d] = %4.2e\n", ii, PseudoPrior[ii]);
   printf ("\n");
#endif

   /* Initialize marker prob distribution to uniform. Change if               */
   /* one marker locus is to be favoured                                      */
   for (ii = 1; ii <= nmrkloc; ii++)
      mrk_distr[ii] = 1.0 / nmrkloc;

   /* Re-Initialize counters */
   rao_freq = 0;
   for (ii = 0; ii < NTPos1; ii++)
   {
      tposit_freq[ii] = 0;
      raoI_est[ii] = 0.0;
   }

   for (iteration = 1; iteration <= NIt; iteration++)
   {
      if (!(iteration % ChkIt))
      {
         printf (" Starting main iteration %d\n", iteration);
         fflush (stdout);
      }

/* +------------------------------------------------+                     */
/* | UPDATE: S's WITHOUT A CHANGE OF TRAIT POSITION |                     */
/* +------------------------------------------------+                     */

      update_loci (LSamplProb, nmrkloc, pedsize, y, nsw, prob, ngenes);

      if (previous_realiz.intvl == 0 && previous_realiz.rec_lft[FEM] == 0.5)
         /* Trait locus is presently unlinked */

/*   +--------------------------------------------------------------+  */
/*   | UPDATE: PROPOSE MOVING UNLINKED TRAIT LOCUS TO LINKED STATE  |  */
/*   +--------------------------------------------------------------+  */

         update_unlnkd_2_lnkd (nmrkloc, pedsize, y, prob, ngenes,
                               n_compnts, p_linked);

      else                              /* Trait locus is presently linked */
         /*   +-------------------------------------------------------------+ */
         /*   |  UPDATE:  PROPOSE MOVING TRAIT LOCUS ALONG THE CHROMOSOME   | */
         /*   |           1. Move may be within marker interval             | */
         /*   |           2. Move may be between marker interval            | */
         /*   |           3. Move may be to UNLINKED STATE                  | */
         /*   +-------------------------------------------------------------+ */
         update_position (p_linked, WindWdth, nmrkloc, pedsize, y, prob, ngenes,
                          n_compnts);

      if ((iteration % ScorIt) == 0)    /*  Calculate Lange estimator */
      {
         rao_freq++;
         calc_raoI (nmrkloc, rao_freq, rao_realiz);
         if (OutRB)
            write_rb (oRBF, &rao_realiz);
      }
      if (SeqImpIt && !(iteration % SeqImpIt))
         mh_restart (nmrkloc, pedsize, prob, y, n_compnts, ngenes);

   }                                    /* END MCMC SAMPLER FOR LOOP */
   printf ("\n");
   prnt_mcmcend ();

   prntLodSHead ();

   prntLodSPos0 ();                     /* unlinked trait position  */
   printf ("%11.6f", PseudoPrior[0]);
   printf ("%9d", tposit_freq[0]);
   printf ("%10s%10s\n", "NA", "NA");

   for (i = 0, ii = 1; i < NTPos + NMarks; ++i)
   {
      if (!prntLodSPos (i))
         printf ("%11s%9s%10s%10s", "NA", "NA", "NA", "NA");  /* marker pos   */
      else
      {
         printf ("%11.6f", PseudoPrior[ii]);
         printf ("%9d", tposit_freq[ii]);
         if (tposit_freq[ii] == 0)      /* Test for test positions which are  */
            printf ("%10s", "NA");      /* not being visited  */
         else if (tposit_freq[0] == 0)  /* Unlinked case not visited */
            printf ("%10s", "INF");
         else
            printf ("%10.4f", log10 ((tposit_freq[ii] * PseudoPrior[0])
                                     / (tposit_freq[0] * PseudoPrior[ii])));
         printf ("%10.4f", log10 (raoI_est[ii] * PseudoPrior[0]
                                  / (raoI_est[0] * PseudoPrior[ii])));
         ++ii;
      }
      printf ("\n");
   }
   fflush (stdout);

/* Close output files                                                    */

   if (OutRB)
      fclose (oRBF);

/* Free memory                                                           */
   free_xxgg ();
   locus_peel_free ();
   free_allocate_loci ();
   free_chromo (0);
   free_gen_pen ();
   free_seg_pen (nmrkloc + 1, &prob);
   free_y (&y);
   free_local (nns, pedsize, nmrkloc);
   free_msmpl ();

/* Write new seed to seed file                                           */
   if (SamplSeedsOut)
   {
      getSeeds (SAMPL);
      if (!(oSeedF = open_fil (OSEED)))
         give_up (ERR_qit);
      writSeeds (oSeedF);
      fclose (oSeedF);
   }

   if (qdex)
   {
      free (qdex);
      qdex = NULL;
   }
   if (trtphen)
   {
      free (trtphen);
      trtphen = NULL;
   }

   exit (EXIT_SUCCESS);

}

/*  +-------------------------+   */
/*  |      SUBROUTINES        |   */
/*  +=========================+   */

/* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~                          */
void  assign_xx (int nummrks, int nped)
/* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~                          */

/* Assign xx0 and xx1 in preparation for use in the L and M samplers          */

{

   int   indvl, locus;

   for (locus = 0; locus <= nummrks; locus++)
      for (indvl = 0; indvl <= nped; indvl++)
      {
         xx0[locus][indvl] = previous_realiz.xx0[locus][indvl];
         xx1[locus][indvl] = previous_realiz.xx1[locus][indvl];
      }

   return;
}

/* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~                      */
void  alloc_previous_realiz (int nped, int nummrks)
/* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~                      */

{
   int   ii;

   previous_realiz.xx0 = malloc ((nummrks + 1) * sizeof (int *));
   alloc_chk (WHERE, previous_realiz.xx0);
   previous_realiz.xx1 = malloc ((nummrks + 1) * sizeof (int *));
   alloc_chk (WHERE, previous_realiz.xx1);
   previous_realiz.xx0[0] = malloc ((nped + 1) * (nummrks + 1) * sizeof (int));
   alloc_chk (WHERE, previous_realiz.xx0[0]);
   for (ii = 1; ii <= nummrks; ++ii)
      previous_realiz.xx0[ii] = previous_realiz.xx0[ii - 1] + (nped + 1);

   previous_realiz.xx1[0] = malloc ((nped + 1) * (nummrks + 1) * sizeof (int));
   alloc_chk (WHERE, previous_realiz.xx1[0]);
   for (ii = 1; ii <= nummrks; ++ii)
      previous_realiz.xx1[ii] = previous_realiz.xx1[ii - 1] + (nped + 1);

   return;
}

/** @} */
