/*****************************************************************************
  program name: lm_pval.c

  Program to do posterior IBD probs
  The state (X) consists of segregation indicators

  X determines the IBD pattern among the specified descendant gametes.

  The data, Y, is the multilocus genotypes of observed individuals in the
  pedigree.

  m_pval implemented, based on Feb 2003 Autozyg program lm_auto.c
                                                                EAT Apr 2003
  Modified to use globals identifying affected individuals,
  component info                                                MAM Jun 2003

******************************************************************************/

/**
 * @defgroup lm_pval lm_pval
 * @ingroup autozyg
 * @{
 */

#define PNAM  "lm_pval"

#include "autozyg.h"

int   main (int argc, char *argv[])
{
   int   i, j, k, low, high, maximp, scol, ncol, wcol;
   int   nloc, ntraits, nmarkers, *loc_flags;
   int ***y;                            /* full genotypic array           */
   double **quant = NULL;
   int **dodo = NULL;
   int   t, c, h, err;
   int   jj;
   int   ngenes = 0, nfou;
   int   nsw, sx, *nns;
   int  *perm_sw, *perm_loc;            /* perm arrays of switches & loci */
   int ***recounts;                     /* [nc][nl][6] counts of 
                                         *             recombinations     */
   double **tot_log_prob_seg;
   double **tot_log_prob_pen;
   double **temp_log_prob_seg;
   double **prob;

   int ***hist;                         /* [nc][nl][NAffInd[c]]
                                         * histogram of no. imputations
                                         * at all loci                    */

   int   scmark[12];                    /* locus counts of exceeding null */

   int   tot_count = 0;                 /* score by component and locus   */

   FILE *pedF, *oSeedF;

   atexit (print_tail);
   PCleanUp = clean_auto;
   atexit (clean_morg);
   print_head (*argv);

   if (!proc_pars (argc, argv, PNAM))   /* process input parameter statements */
      exit (EXIT_FAILURE);

   if (!proc_auto ())                   /* check statements for completeness  */
      exit (EXIT_FAILURE);              /* and consistency
                                         * and prepare controls for lm_auto   */

   if (!proc_seeds ())                  /* set up seed parameters             */
      exit (EXIT_FAILURE);

   if (!proc_fils ())                   /* set up pedigree file parameters    */
      exit (EXIT_FAILURE);

   setSeeds (SAMPL);                    /* initializes seeds for rans()       */

   if (!chk_pedfil ())                  /* checks pedigee file for errors     */
      give_up (ERR_qit);

   if (!mak_opedfil ())                 /* writes new pedigree file,          */
      give_up (ERR_qit);                /* if there are changes               */

   if (!(pedF = open_ped ()))           /* reopens pedigree file              */
      give_up (ERR_qit);

   alloc_ped ();
   input_ped (pedF, 0, dodo, quant);    /*input pedigree and trait data       */

   nloc = NLoci - 1;

   nfou = count_founders ();

   MaxNsw = 2 * (pedsize - nfou);
   nns = malloc ((n_compnts + 1) * sizeof (int));
   alloc_chk (WHERE, nns);

   loc_flags = malloc ((nloc + 1) * sizeof (int));
   alloc_chk (WHERE, loc_flags);

/* set up trait structures */
/* read (1) or set (0) information about the Traits (including markers)   */

   ntraits = set_traits (NamdLoc, nloc, loc_flags, pedF);
   nmarkers = print_traits (ntraits, loc_flags);

   printf (" \n");
   fflush (stdout);

/* now allocate space for genotypes y for our values of pedsize, nloc   */

   y = alloc_y (nmarkers, pedsize);

   if (TraitPheno)
   {
      qdex = calloc (pedsize + 1, sizeof (int));
      alloc_chk (WHERE, qdex);
      trtphen = calloc (pedsize + 1, sizeof (int));
      alloc_chk (WHERE, trtphen);
   }

   /* now set marker data on observed individuals */
   /* marker loci are labelled 1,..., nloc */

#if 0
   if (nmarkers < nloc)
   {
      set_markmap (nloc, nmarkers, loc_flags);
      reset_traitmap (nloc, nmarkers, loc_flags);  /* lm_auto only--
                                                      reset Loc0, Rec0 */
      nloc = nmarkers;
      NLoci = nloc + 1;
   }
#endif

   set_markdat (y, nloc, n_compnts);

   /* set data for the trait */

   if (AffTrIn)
      set_traitdat (y, qdex, trtphen, pedsize, n_compnts);

   set_pedaff (y, n_compnts);

   hist = alloc_affhist ();

   alloc_xxgg (nloc, pedsize);

   /* initialize seeds for rans() */
   setSeeds (SAMPL);

   /* now set and index switches */

   nsw = list_switches (nloc, nns);
   printf (" Number of meiosis switches to sample is %d\n", nsw);
   fflush (stdout);

   /* allocate and set arrays of marker and trait positions 
    * and recombination fractions  */

   alloc_chromo (nloc, 0);

   for (sx = FEM; sx <= MAL; sx++)
      set_chromosomes (nloc, Marek[sx], 0, &Loc0, &Rec0[sx], sx);

   free_marek ();

   /* set up permutation arrays for loci and switches  if Scan (not Step) */

   if (Scan)
   {
      perm_sw = malloc (nsw * sizeof (int));
      alloc_chk (WHERE, perm_sw);
      for (i = 0; i < nsw; ++i)
         perm_sw[i] = i;
      perm_loc = malloc ((nloc + 1) * sizeof (int));
      alloc_chk (WHERE, perm_loc);
      for (i = 0; i <= nloc; ++i)
         perm_loc[i] = i;
   }

   /* set up recombination patterns and initial probs */

   /* Allocate locus structures    -moved to main program, EAT June 2003 */
   /* if TraitPheno  penetrance will be set to incomplete, EAT 12/2003   */
   allocate_loci (nloc + 1, NAll, Qq, pedF, 1);

   /* Use L-sampler or sequential imputation for set_up and initial xx0, xx1
      (maternal and paternal segregation indicators) */
   if (SmplSetup == LSMPL)              /* use L-sampler for set-up  */
      set_lsmpl (0, nloc, y, TraitPheno);

   else                                 /* use sequential imputation (AWG) */
      seq_imp (NSeqImp, pedsize, nloc + 1, y, (double *) 0, 1.0);

   peel_trace = TRACE_LEVEL_0;          /* Turn on trace if level >0 */

   alloc_loglik (&recounts, &tot_log_prob_seg, &temp_log_prob_seg,
                 &tot_log_prob_pen);

   /* set genes from initial switches */
   /* Allocate memory for gen_pen()  for each trait locus  */
   /*  alloc_seg_pen called ONCE ONLY for all markers      */
   /*     alloc of prob array for seg_pen results added here EAT 2/2002 */

   k = 0;
   for (j = 0; j <= nloc; ++j)
   {
      ngenes = set_genes (j);

      if (loci[j].locus_type == TRAIT_LOCUS)
         alloc_gen_pen (ngenes, j);
      else if (!k++)
         alloc_seg_pen (ngenes, nloc + 1, y, &prob);
   }
   for (j = 0; j <= nloc; ++j)
      for (c = 0; c < n_compnts; ++c)
         prob[c][j] = seg_pen2 (c, y, Qq[j], j, ngenes, &err);

   /* probability now computed separately by component  */

   for (c = 0; c < n_compnts; ++c)
   {
      for (h = 0; h <= nloc; ++h)
      {
         tot_log_prob_seg[c][h] = 0.0;
         tot_log_prob_pen[c][h] = 0.0;
      }
      for (j = 0; j <= nloc; ++j)
      {
         for (jj = 0; jj < 6; ++jj)
            recounts[c][j][jj] = 0;
      }
   }

   /* count total number of patterns */

   alloc_logprob ();

   alloc_msmpl ();

   prnt_mcmcstrt ();

   /* now do burn-in iterations */

   printf (" Now starting %d iterations of burn-in\n", NBrnIt);
   for (t = 1; t <= NBrnIt; t++)
   {
      if (!(t % ChkIt))
      {
         printf (" Starting burnin iteration %d\n", t);
         fflush (stdout);
      }
      sampling (t, 0, nloc, nsw, ngenes, perm_loc, perm_sw, y, prob, 0);
   }

   /* now do MCMC sampling, and score results */

   printf (" Now starting %d main MCMC iterations of burn-in\n", NIt);
   fflush (stdout);
   for (t = 0; t < NIt; ++t)
   {
      sampling (t, 0, nloc, nsw, ngenes, perm_loc, perm_sw, y, prob, 0);
      /* MCMC sampling step or scan */

      if (!((t + 1) % ScorIt))
      {
         score_aff (hist);              /* scores for affecteds               */

         /* this is scoring of log-segs and log-pens */
         for (c = 0; c < n_compnts; ++c)
         {

            full_seg_prob (nloc, c, recounts, temp_log_prob_seg);
            for (j = 0; j < nloc; ++j)
            {
               tot_log_prob_pen[c][j] += log (prob[c][j]);
               tot_log_prob_seg[c][j] += temp_log_prob_seg[c][j];
            }
            tot_log_prob_pen[c][nloc] += log (prob[c][nloc]);
         }
         ++tot_count;
      }
   }
   printf ("\n");
   prnt_mcmcend ();

#if 0
   printf ("\n Total count = %d\n\n", tot_count);
   printf (" Scoring of exceedence counts at markers \n");
   for (i = 1; i < NLoci; ++i)
      printf (" %d", scmark[i]);
   printf ("\n\n");
   printf (" Scoring of counts at each locus \n\n");
   printf ("   null locus:");
   for (i = 0; i < naff; ++i)
      printf ("  %d", hist[0][i]);
   for (j = 1; j < NLoci; ++j)
   {
      printf ("\n   locus%3d:  ", j);
      for (i = 0; i < naff; ++i)
         printf ("  %d", hist[j][i]);
   }
   printf ("\n\n");
#endif

   for (c = 0; c < n_compnts; ++c)
   {
      pr_cmpnthead (c);

      if (!NAffInd[c])
         printf ("    There are no affected individuals for this"
                 " component\n\n");
      else
      {
         printf ("    %d affected individuals\n\n", NAffInd[c]);

         low = NAffInd[c];
         high = 0;
         maximp = 0;
         for (j = 0; j < NLoci; ++j)
            for (k = 0; k < NAffInd[c]; ++k)
            {
               if (hist[c][j][k] > 0 && k < low)
                  low = k;
               if (hist[c][j][k] > 0 && k > high)
                  high = k;
               if (hist[c][j][k] > maximp)
                  maximp = hist[c][j][k];
            }

         scol = 7 + LocNamLen + 1;
         get_colfmt (80 - scol - 1, intWid (0, high), intWid (0, maximp),
                     1, &ncol, &wcol);

         printf ("    Number of imputations with values:\n");

         for (i = 0, k = low; k <= high; ++i, ++k)
         {
            if (!(i % ncol))
               printf ("\n%*s ", scol, " ");
            printf (" %*d", wcol, k + 1);
         }
         printf ("\n");

         for (j = 0; j < NLoci; ++j)
         {
            printf ("\n       %*s: ", LocNamLen, traits[chromo[0][j]].self);
            for (i = 0, k = low; k <= high; ++i, ++k)
            {
               if (i && !(i % ncol))
                  printf ("\n%*s ", scol, " ");
               printf (" %*d", wcol, hist[c][j][k]);
            }
         }
         printf ("\n\n");
      }

      pr_log_prob_seg (1, nloc, -nns[c + 1], tot_log_prob_seg[c], tot_count);

      pr_log_prob_pen (1, nloc, tot_log_prob_pen[c], tot_count);

      pr_reco_counts (1, nloc, recounts[c]);

#if 0
      printf ("    Average total log-probability of segregations:\n");
      printf ("       first locus (constant) %8.3f\n       by map interval:",
              -nns[c + 1] * log (2.0));

      for (j = 0; j < nloc; ++j)
      {
         if (!(j % 8))
            printf ("\n       ");
         printf (" %8.3f", tot_log_prob_seg[c][j] / tot_count);
      }
      printf ("\n\n");

      printf ("    Average total log-probability of penetrances,"
              " by locus:");
      for (j = 1; j <= nloc; ++j)       /* null unlinked locus ignored here   */
      {
         if (!((j - 1) % 8))
            printf ("\n      ");
         printf (" mark%2d  ", chromo[0][j]);
      }
      for (j = 1; j <= nloc; ++j)
      {
         if (!((j - 1) % 8))
            printf ("\n    ");
         printf (" %8.3f", tot_log_prob_pen[c][chromo[0][j]] / tot_count);
      }
      printf ("\n\n");

      /* print out recombination counts  */
      printf ("    Recombination counts for map intervals:\n");
      printf ("       intv   M: total    recomb   non-rec "
              "  F: total    recomb   non-rec\n");
      for (j = 0; j < nloc; ++j)
      {
         printf ("       %4d ", j + 1);
         for (i = 0; i < 6; ++i)
         {
            if (i == 3)
               printf (" ");
            printf ("%10d", recounts[c][j][i]);
         }
         printf ("\n");
      }
      printf ("\n");
      fflush (stdout);
#endif
   }
   printf ("\n ");                      /* ****  now we are done **** */
   fflush (stdout);

   /* Free up allocated memory */

   free_xxgg ();
   free_y (&y);

   free_loglik (&recounts, &tot_log_prob_seg, &temp_log_prob_seg,
                &tot_log_prob_pen);
   locus_peel_free ();
   free_allocate_loci ();
   free_chromo (0);
   free_gen_pen ();
   free_seg_pen (nloc + 1, &prob);
   free_logprob ();
   free_msmpl ();
   free_traits ();

   if (SamplSeedsOut)
   {
      getSeeds (SAMPL);
      if (!(oSeedF = open_fil (OSEED)))
         give_up (ERR_qit);
      writSeeds (oSeedF);
      fclose (oSeedF);
   }

   if (nns)
   {
      free (nns);
      nns = NULL;
   }
   if (loc_flags)
   {
      free (loc_flags);
      loc_flags = NULL;
   }
   if (Scan)
   {
      if (perm_sw)
      {
         free (perm_sw);
         perm_sw = NULL;
      }
      if (perm_loc)
      {
         free (perm_loc);
         perm_loc = NULL;
      }
   }
   if (qdex)
   {
      free (qdex);
      qdex = NULL;
   }
   if (trtphen)
   {
      free (trtphen);
      trtphen = NULL;
   }

   exit (EXIT_SUCCESS);
}

/** @} */
