
#include "quant.h"
#include "peelie.h"

/*************************************************************************
  
  program name: mjrpeel.c
  
  program to calculate log-likelihood on zero-loop pedigrees, 
      for a major Mendelian locus by the method of peeling.
  
  Data are quantitative trait values with any fixed effects, 
      and/or polygenic values subtracted.
  
  Program from SWG, from original peelie.c program; date: 10/30/90
  
  Modified for major gene peeling; EAT; 7/29/92
  
  Includes also penetrance routines, peripheral parent peeling,
     and routines to retrieve stored R-functions.;  EAT ; 7/29/92
  
  Includes also a flag to indicate inclusion of peeling at 2 linked loci.
     and option to peel linked loci;   EAT; 8/25/92
  
  Modified to permit use with multi-allele marker programs; SL; 1/93.
  
  This version installed in quant library; EAT; 1/93.
  Modified for fractional penetrances; EAT; 10/93.
  
  Modified periph() so that if a founder with a previously stored
      R-Function is peeled, the population genotype frequencies  
      are incorporated properly; SCH; 5/96
 
  (EAT) Will now peel several pedigrees, but does assume that peeling
      sequence does each pedigree in turn. Final pivot of each pedigree
      should be 0, and the final pivot of final pedigree should be -1.
         3/22/97
  
  (EAT) Modified to use peeling sequence from integer array, instead
      of reading from file.  3/25/97
    
  (EAT) Corrected so that final pivots of compoenents and of data sets
      are now -1 and -2 instead of 0 and -1, 5/22/97.
 
**************************************************************************/

double pengen (double yy, int kk, double *g_mu, double pop, double lam, int im)
{
   double tmp;
   if (im < 2)
      tmp = yy - g_mu[mjrgene[kk]];
   else
      tmp = yy - lam * g_mu[mjrgene[kk]];
   tmp = -0.5 * tmp * tmp / SigErr;
   tmp = exp (tmp) * pop;
   return (tmp);
}

/**************************************************************************/

void  prev_rfn (double *val, int k, int rw)
{
   int   i;
   for (i = 0; i < k; ++i)
      val[i] = qstore[i][rw];
}

/**************************************************************************/

void  periph (double *ym, double *g_mu, double *kk, int pp, double pop,
              int mflag, double lam, int im)
{
   int   j, obs, mkr;

   if (istore[pp] != 0)
   {
      istore[pp] = 0;

      /*
       * Ensure that founders with a previously stored R-Function get 
       * the founder genotype frequencies taken into account  - SCH 5/96 
       */

      if (isNullNode (nghd[pp].pa))
         for (j = 0; j < n_geno; ++j)
            kk[j] = qstore[j][pp] * geno_freq[j];
      else
         for (j = 0; j < n_geno; ++j)
            kk[j] = qstore[j][pp];
   }
   else
      for (j = 0; j < n_geno; ++j)
         kk[j] = geno_freq[j];

   if (idatu[pp] == 0)
   {
      obs = qdex[pp];
      ++idatu[pp];

      if (obs > 0)
      {
         for (j = 0; j < n_geno; ++j)
            kk[j] *= pengen (ym[obs], j, g_mu, pop, lam, im);
      }
      if (mflag > 0)
      {
         mkr = h[pp];
         if (mkr != MISSING)
         {
            for (j = 0; j < n_geno; ++j)
               kk[j] *= mkpen[mkgeno[j]][mkr];
         }
      }
   }
}
/**************************************************************************/

double mjrpeel (double *ym, double *g_mu, int *peelseq, int mflag,
                double lam, int im)
{
   int   nch, npeelch, ipivot, kid, nt, ma, pa, child[NS], obs;
   double *qval, *mval, *pval;
   double cc, prob, logprob, pop;
   double tt, c[21][21];
   int   i, j, k, i1, i2, psq, marker;

   qval = (double *) malloc (n_geno * sizeof (double));
   mval = (double *) malloc (n_geno * sizeof (double));
   pval = (double *) malloc (n_geno * sizeof (double));
   /*
    * c = (double **) malloc(n_geno * sizeof(double *)); 
    * c[0] = (double *) malloc(n_geno*n_geno*sizeof(double)); 
    * for (i = 1; i < n_geno; ++i)
    *     c[i] = c[i-1] + i * n_geno; 
    */

   logprob = 0.0;
   pop = 1.0 / (sqrt (PPI * SigErr));
   ipivot = 0;
   for (i1 = 0; i1 < pedsize; ++i1)
      istore[i1] = idatu[i1] = 0;

   psq = 0;

   /*
    * printf ("\n now set to peel"); fflush(stdout);     
    */

   while (ipivot > -2)                  /* for final pedigree, 
                                         * final pivot = -2      */
   {
      nt = 2;
      prob = 0.0;

      /* for each sibship of the pedigree in turn */
      while (nt > 1)
      {

         for (i1 = 0; i1 < n_geno; ++i1)
            for (i2 = 0; i2 < n_geno; ++i2)
               c[i1][i2] = 1.0;

         pa = peelseq[psq];
         ma = peelseq[psq + 1];
         ipivot = peelseq[psq + 2];
         nch = peelseq[psq + 3];

         /*
          * printf ("\n new family pa= %d, ma=%d, pivot=%d \n",
          * pa ,ma ,ipivot);    
          */
         for (i2 = 1; i2 <= nch; ++i2)
            child[i2] = peelseq[psq + 3 + i2];
         psq += 4 + nch;

         /* printf ("\n new family; pivot = %d", ipivot);   */
         fflush (stdout);

         npeelch = 0;
         nt = 3;

         for (i1 = 1; i1 <= nch; ++i1)
         {
            if (ipivot != child[i1])
            {
               kid = child[i1];

               /* incorporate any previous likelihood on child  */

               if (istore[kid] != 0)
               {
                  istore[kid] = 0;
                  prev_rfn (qval, n_geno, kid);
               }
               else
               {
                  for (i = 0; i < n_geno; ++i)
                     qval[i] = 1.0;
               }

               obs = qdex[kid];
               marker = h[kid];
               ++idatu[kid];

               for (k = 0; k < n_geno; k++)
               {
                  if (obs > 0)
                     qval[k] *= pengen (ym[obs], k, g_mu, pop, lam, im);
                  if ((marker != MISSING) && (mflag > 0))
                     qval[k] *= mkpen[mkgeno[k]][marker];
               }

               for (i = 0; i < n_geno; ++i)
               {
                  for (j = 0; j < n_geno; ++j)
                  {
                     tt = 0.0;
                     for (k = 0; k < n_geno; ++k)
                        tt += qval[k] * trans[i][j][k];
                     c[i][j] *= tt;
                  }
               }
               ++npeelch;
            }
         }
         if (npeelch == nch)
         {
            if (ma != ipivot)
            {
               periph (ym, g_mu, mval, ma, pop, mflag, lam, im);
               --nt;
               for (i = 0; i < n_geno; ++i)
               {
                  qval[i] = 0.0;
                  for (j = 0; j < n_geno; ++j)
                     qval[i] += mval[j] * c[i][j];
               }
            }
            if (pa != ipivot && ma == ipivot)
            {
               periph (ym, g_mu, pval, pa, pop, mflag, lam, im);
               --nt;
               for (i = 0; i < n_geno; ++i)
               {
                  qval[i] = 0.0;
                  for (j = 0; j < n_geno; ++j)
                  {
                     qval[i] += pval[j] * c[j][i];
                  }
               }
            }
            /* peeling of pedigree completed  */
            else if (pa != ipivot && ma != ipivot)
            {
               periph (ym, g_mu, pval, pa, pop, mflag, lam, im);
               --nt;

               for (i = 0; i < n_geno; ++i)
                  prob += qval[i] * pval[i];
            }
         }
         if (npeelch < nch)
         {
            periph (ym, g_mu, mval, ma, pop, mflag, lam, im);
            periph (ym, g_mu, pval, pa, pop, mflag, lam, im);

            for (i = 0; i < n_geno; ++i)
            {
               qval[i] = 0.0;

               for (j = 0; j < n_geno; ++j)
               {
                  for (k = 0; k < n_geno; ++k)
                  {
                     qval[i] += pval[j] * mval[k] * c[j][k] * trans[j][k][i];
                  }
               }
            }
         }
         if (ipivot > -1)
         {
            /* printf ("\n family complete ");    */

            if (istore[ipivot] != 0)
            {
               istore[ipivot] = 0;
               for (i = 0; i < n_geno; ++i)
                  qval[i] *= qstore[i][ipivot];
            }
            cc = qval[0];
            for (i = 1; i < n_geno; ++i)
            {
               if (cc < qval[i])
                  cc = qval[i];
            }
            logprob += log (cc);
            /* printf (";  current logprob %f", logprob);   */
            for (i = 0; i < n_geno; ++i)
               qval[i] /= cc;

            istore[ipivot] = 1;
            for (i = 0; i < n_geno; ++i)
               qstore[i][ipivot] = qval[i];
         }
      }
      logprob += log (prob);
   }
   for (i1 = 0; i1 < pedsize; ++i1)
   {
      if (istore[i1] != 0)
         printf ("\n Unused likelihood on %s still stored\n", nghd[i1].self);
      if (idatu[i1] != 1)
         printf ("\n Data on %s used %d times\n", nghd[i1].self, idatu[i1]);
   }
   free ((void *) pval);
   free ((void *) mval);
   free ((void *) qval);

   /* printf ("\n Final logprob %f\n",logprob);      */

   return (logprob);
}
