/****************************************************************************
 *                                                                          *
 *     Loki - Programs for genetic analysis of complex traits using MCMC    *
 *                                                                          *
 *             Simon Heath - University of Washington                       *
 *                                                                          *
 *                        August 1997                                       *
 *                                                                          *
 * peel_util.c:                                                             *
 *                                                                          *
 ****************************************************************************/

#include "lsample.h"

int   num_bits (const int n_all)
{
   int   nb;

   nb = (int) (log ((double) n_all) / log (2.0));
   if ((1 << nb) < n_all)
      nb++;
   return nb;
}

/* General abort routine in case of error */
void  abt (const char *file, const int line, const char *fmt,...)
{
   va_list args;

   if (stdout)
      (void) fflush (stdout);
   (void) fprintf (stderr, "[%s:%d] ", file, line);
   va_start (args, fmt);
   (void) vfprintf (stderr, fmt, args);
   va_end (args);
   _exit (EXIT_FAILURE);
}

/* Print out allele for ind. xx as xxm (maternal) or xxp (paternal) */
void  print_orig_allele_id (FILE * fptr, const int i)
{
   (void) fprintf (fptr, "%s%c", nghd[abs (i) - 1].self, i > 0 ? 'm' : 'p');
}

void  print_simple_element (FILE * fptr, const struct Simple_Element *simple_em,
                            struct R_Func *rf)
{
   Indlist *id_list;
   int   i;

   fprintf (fptr, "%s,%s ", simple_em->sire->self, simple_em->dam->self);
   id_list = simple_em->off;
   i = 0;
   while (id_list)
   {
      fprintf (fptr, "%c%s", i++ ? ',' : '[', id_list->current->self);
      id_list = id_list->next;
   }
   if (i)
      fputs ("] ", fptr);
   if (simple_em->pivot)
      fprintf (fptr, " -> %s", simple_em->pivot->self);
   if (simple_em->pivot1)
      fprintf (fptr, ",%s", simple_em->pivot1->self);
   for (i = 0; i < simple_em->n_rfuncs; i++)
   {
      fprintf (fptr, " (%s)",
               nghd[rf[simple_em->index[i]].id_list[0] - 1].self);
   }
   return;
}

void  print_complex_element (FILE * fptr,
                             const struct Complex_Element *complex_em,
                             struct R_Func *rf)
{
   struct R_Func *rf1;
   int   i, j;

   for (i = 0; i < complex_em->n_peel; i++)
   {
      if (i)
         fputc (',', fptr);
      print_orig_allele_id (fptr, complex_em->involved[i]);
   }
   if (i < complex_em->n_involved)
      for (; i < complex_em->n_involved; i++)
      {
         if (i == complex_em->n_peel)
            fputs (" -> ", fptr);
         else
            fputc (',', fptr);
         print_orig_allele_id (fptr, complex_em->involved[i]);
      }
   for (i = 0; i < complex_em->n_rfuncs; i++)
   {
      fputc (' ', fptr);
      rf1 = rf + complex_em->index[i];
      for (j = 0; j < rf1->n_ind; j++)
      {
         fputc (j ? ',' : '(', fptr);
         print_orig_allele_id (fptr, complex_em->involved[rf1->id_list[j]]);
      }
      fputc (')', fptr);
   }
   return;
}

/* Print out peeling sequence */
void  print_sequence (FILE * fptr, struct Peelseq_Head *pp, struct R_Func *rf)
{
   struct Complex_Element *complex_em;
   struct Simple_Element *simple_em;

   while (pp && pp->type)
   {
      if (pp->type == PEEL_SIMPLE)
      {
         simple_em = pp->ptr.simple;
         fputs ("(S) ", fptr);
         print_simple_element (fptr, simple_em, rf);
         fputc ('\n', fptr);
         pp = &simple_em->next;
      }
      else
      {
         complex_em = pp->ptr.complex;
         fputs ("(C) ", fptr);
         print_complex_element (fptr, complex_em, rf);
         fputc ('\n', fptr);
         pp = &complex_em->next;
      }
   }
   return;
}

void  free_peelseq (struct Peelseq_Head *pp)
{
   struct Simple_Element *simple_em;
   struct Complex_Element *complex_em;
   void *old_ptr = 0;

   while (pp->type)
   {
      if (pp->type == PEEL_SIMPLE)
      {
         simple_em = pp->ptr.simple;
         if (old_ptr)
            free (old_ptr);
         pp = &simple_em->next;
         free (simple_em->off);
         if (simple_em->n_rfuncs)
            free (simple_em->index);
         old_ptr = (void *) simple_em;
      }
      else
      {
         complex_em = pp->ptr.complex;
         if (old_ptr)
            free (old_ptr);
         pp = &complex_em->next;
         free (complex_em->involved);
         old_ptr = (void *) complex_em;
      }
   }
   if (old_ptr)
      free (old_ptr);
   return;
}
