/****************************************************************************
 *                                                                          *
 *    Penetrance routines moved from lsample_peel.c: EAT April 2002         *
 *                                                                          *
 ****************************************************************************/

#include "lsample.h"

static double *res, sige, konst;

void  set_qtlpen (double *ym, double ve)
/* sets up static data vector and constants for qtlpen  */
{
   res = ym;
   sige = ve;
   konst = 1.0 / sqrt (PPI * sige);
}

/* Change so that we are multiplying the input vector by the penetrance 
 * (we just zero out the impossible terms in this case)
 * Also change so we get the number of alleles from the locus structure
 * rather than the global n_all, allowing us to use the same routine with
 * both L and M samplers - SCH June 2001 */
void  recessive (double *val, const Ind * id, const struct Locus *locus,
                 double lam)
{
   int   obs, k;
   if (locus->n_all != 2)
      printf ("\n WARNING: trait locus has %d alleles\n", locus->n_all);
   obs = qdex[id->index];
   if (obs > 0)
   {
      k = trtphen[id->index];
      if (k == 4)
      {
         val[0] = 0.0;
         val[1] = 0.0;
         val[2] = 0.0;
      }
      else if (k == 3)
      {
         val[0] = 0.0;
         val[3] = 0.0;
      }
      else if (k == 2)
      {
         val[3] = 0.0;
      }
      else if (k == 1)
      {
         val[1] = 0.0;
         val[2] = 0.0;
         val[3] = 0.0;
      }
      else
         printf ("\n UNRECOGNIZED TRAIT PHENOTYPE of %s is %d",
                 id->self, k);
   }
   return;
}

/* Added by AWG */
void  incomplete (double *val, const Ind * id, const struct Locus *locus,
                  double lam)
{
   int   obs, k;
   if (locus->n_all != 2)
      printf ("\n WARNING: trait locus has %d alleles\n", locus->n_all);

   obs = qdex[id->index];
   if (obs > 0)
   {
      k = trtphen[id->index];

      /*  
         0    - homozygous for the non disease allele
         1, 2 - heterozygous: carrier of the disease allele
         3    - homozygous for the disease allele
       */

      if (k == 2)                       /* Affected */
      {
         val[0] = locus->g_mu[0] * val[0];
         val[1] = locus->g_mu[1] * val[1];
         val[2] = locus->g_mu[1] * val[2];
         val[3] = locus->g_mu[2] * val[3];
      }
      else if (k == 1)                  /* Unaffected */
      {
         val[0] = (1 - locus->g_mu[0]) * val[0];
         val[1] = (1 - locus->g_mu[1]) * val[1];
         val[2] = (1 - locus->g_mu[1]) * val[2];
         val[3] = (1 - locus->g_mu[2]) * val[3];
      }
      else
         printf ("\n UNRECOGNIZED TRAIT PHENOTYPE of %s is %d",
                 id->self, k);

   }

   return;
}

void  ordinal (double *val, const Ind * id, const struct Locus *locus,
               double lam)
{
   /* This penetrance function was used in the analysis of the GAW13 */
   /* data. The ordinal trait contains 5 levels where 1 corresponds  */
   /* to low values of the trait. The penetrance values are based    */
   /* on estimates from Bill's EM algorithm which fits a sinlge      */
   /* major gene.  AWG 07/31/02                                      */
   /* Note:  
      0    - homozygous for the non disease allele
      1, 2 - heterozygous: carrier of the disease allele
      3    - homozygous for the disease allele
    */

   int   obs, k;
   if (locus->n_all != 2)
      printf ("\n WARNING: trait locus has %d alleles\n", locus->n_all);
   obs = qdex[id->index];
   if (obs > 0)
   {
      k = trtphen[id->index];
      if (k == 5)
      {
         val[0] = 0.15 * val[0];
         val[1] = 0.15 * val[1];
         val[2] = 0.15 * val[2];
         val[3] = 0.95 * val[3];
      }
      else if (k == 4)
      {
         val[0] = 0.22 * val[0];
         val[1] = 0.22 * val[1];
         val[2] = 0.22 * val[2];
         val[3] = 0.05 * val[3];
      }
      else if (k == 3)
      {
         val[0] = 0.22 * val[0];
         val[1] = 0.22 * val[1];
         val[2] = 0.22 * val[2];
         val[3] = 0.0 * val[3];
      }
      else if (k == 2)
      {
         val[0] = 0.20 * val[0];
         val[1] = 0.20 * val[1];
         val[2] = 0.20 * val[2];
         val[3] = 0.0 * val[3];
      }
      else if (k == 1)
      {
         val[0] = 0.21 * val[0];
         val[1] = 0.21 * val[1];
         val[2] = 0.21 * val[2];
         val[3] = 0.0 * val[3];
      }
      else
         printf ("\n UNRECOGNIZED TRAIT PHENOTYPE of %s is %d",
                 id->self, k);
   }
   return;
}

void  liability (double *val, const Ind * id, const struct Locus *locus,
                 double lam)
{
   /*  Used in the analysis of the VGEOAD data. AWG */

   int   obs, k, j;
   int   ncols, nrows;
   double **penmat;                         /* Penetrance matrix - rows = liability classes , columns = dd Dd  dD DD where dd is undiseased */
   /* penmat[ncols][nrows] i.e. penmat[0,..][0,..]  */

   ncols = 4;                           /* number of genotypic classes */
   nrows = 10;                          /* number of liability classes */
   penmat = (double **) malloc ((ncols) * sizeof (double *));
   alloc_chk (WHERE, penmat);
   penmat[0] = (double *) malloc (nrows * ncols * sizeof (double));
   for (j = 1; j < ncols; j++)
      penmat[j] = penmat[j - 1] + nrows;

   /* penmat assignment (i.e we are modelling sporadics, that is why penmat[0][.] is nonzero */
   penmat[0][0] = 0.0002;
   penmat[1][0] = 0.05;
   penmat[2][0] = 0.05;
   penmat[3][0] = 0.05;                 /* liability class 1 */
   penmat[0][1] = 0.0057;
   penmat[1][1] = 0.15;
   penmat[2][1] = 0.15;
   penmat[3][1] = 0.15;                 /* liability class 2 */
   penmat[0][2] = 0.0124;
   penmat[1][2] = 0.275;
   penmat[2][2] = 0.275;
   penmat[3][2] = 0.275;                /* liability class 3 */
   penmat[0][3] = 0.024;
   penmat[1][3] = 0.425;
   penmat[2][3] = 0.425;
   penmat[3][3] = 0.425;                /* liability class 4 */
   penmat[0][4] = 0.0435;
   penmat[1][4] = 0.575;
   penmat[2][4] = 0.575;
   penmat[3][4] = 0.575;                /* liability class 5 */
   penmat[0][5] = 0.0846;
   penmat[1][5] = 0.725;
   penmat[2][5] = 0.725;
   penmat[3][5] = 0.725;                /* liability class 6 */
   penmat[0][6] = 0.132;
   penmat[1][6] = 0.825;
   penmat[2][6] = 0.825;
   penmat[3][6] = 0.825;                /* liability class 7 */
   penmat[0][7] = 0.187;
   penmat[1][7] = 0.875;
   penmat[2][7] = 0.875;
   penmat[3][7] = 0.875;                /* liability class 8 */
   penmat[0][8] = 0.299;
   penmat[1][8] = 0.925;
   penmat[2][8] = 0.925;
   penmat[3][8] = 0.925;                /* liability class 9 */
   penmat[0][9] = 0.688;
   penmat[1][9] = 0.975;
   penmat[2][9] = 0.975;
   penmat[3][9] = 0.975;                /* liability class 10 */

   if (locus->n_all != 2)
      printf ("\n WARNING: trait locus has %d alleles\n", locus->n_all);
   obs = qdex[id->index];
   if (obs > 0)
   {
      k = trtphen[id->index];

      if (k == 2)                       /* Affected */
      {
         /* The -1 is so that the liability classes now run from 0, ...,  n-1 */
         val[0] = penmat[0][data[id->index][6] - 1] * val[0];
         val[1] = penmat[1][data[id->index][6] - 1] * val[1];
         val[2] = penmat[2][data[id->index][6] - 1] * val[2];
         val[3] = penmat[3][data[id->index][6] - 1] * val[3];
      }

      else if (k == 1)                  /* Unaffected */
      {
         val[0] = (1.0 - penmat[0][data[id->index][6] - 1]) * val[0];
         val[1] = (1.0 - penmat[1][data[id->index][6] - 1]) * val[1];
         val[2] = (1.0 - penmat[2][data[id->index][6] - 1]) * val[2];
         val[3] = (1.0 - penmat[3][data[id->index][6] - 1]) * val[3];
      }
      else
         printf ("\n UNRECOGNIZED TRAIT PHENOTYPE of %s is %d",
                 id->self, k);
   }
   if (penmat)
   {
      free (penmat[0]);
      free (penmat);
      penmat = NULL;
   }

   return;
}

void  qtlpen (double *val, const Ind * id, const struct Locus *locus,
              double lam)
{
   double *g_mu, z, y;
   int   i, j, k, obs;

   if (CHK_PEEL (TRACE_LEVEL_1))
      printf ("Inserting penetrance probabilities for %s\n", id->self);
   obs = qdex[id->index];
   if (obs > 0)
   {
      g_mu = locus->g_mu;
      y = res[obs];
      k = 0;
      for (i = 0; i < locus->n_all; i++)
         for (j = 0; j <= i; j++)
         {
            z = y - lam * g_mu[k];
            z = exp (-.5 * z * z / sige) * konst;
            val[i * locus->n_all + j] *= z;  /* EAT bug fixed, 4/2002 */
            if (i != j)
               val[j * locus->n_all + i] *= z;  /* EAT bug fixed. 4/2002 */
            k++;                        /* EAT fixed, 6/2003  */
         }
   }
}
